<#
    .SYNOPSIS
    Update Orchestrator and Identity Server integration urls.

    .DESCRIPTION
    Update UiPath.Orchestrator.dll.config OpenIdConnect and Identity integration related settings with the new Orchestrator Url and Identity Server Url.
    Update appsettings.Production.json(Identity Server config) with the new values for Orchestrator Url and Identity Server Url.
    Update the [Identity].[ClientRedirectUris] table with the new Orchestrator Url.
    SQLServer PowerShell module is required.
    Use this command with caution as it does not have roll-back mechanism if an error occurs during the invocation.

    .PARAMETER SiteName
    The name of the Orchestrator site that can be found in IIS on the target machine. Usually this can be found as 'UiPath Orchestrator'.

    .PARAMETER OrchestratorUrl
    The new Orchestrator url used by Identity Server in authentication flows.
    
    .PARAMETER IdentityServerUrl
    The new Identity Server url used by Orchestrator in authentication flows.
    If it is not set the default value is '{OrchestratorUrl}/Identity'.

    .PARAMETER SqlConnectionString
    If it is not set the value is read from appsettings.Production.json(Identity Server config).
    The connection string to the UiPath database.
#>

function Update-UiPathUrl {
    [CmdletBinding()]
    Param(
        [Parameter(Mandatory = $true)]
        [string] $SiteName,

        [Parameter(Mandatory = $true)]
        [string] $OrchestratorUrl,

        [Parameter(Mandatory = $false)]
        [string] $IdentityServerUrl,

        [Parameter(Mandatory = $false)]
        [string] $SqlConnectionString
    )

    $orchestratorInstallation = Get-OrchestratorInstallation -SiteName $SiteName
    
    if (-not (Test-OrchestratorLatestVersion -OrchestratorInstallation $orchestratorInstallation)) {
        Write-ErrorMessage "UiPath url can not be updated for an Orchestrator version lower than 20.10."
        return
    }

    $OrchestratorUrl = $OrchestratorUrl.ToLower()

    if (-not $IdentityServerUrl) {
        $IdentityServerUrl = -join ($OrchestratorUrl, "/identity")
        $identitySite = Get-Item IIS:\Sites\$SiteName\Identity
        $identityPath = $identitySite.PhysicalPath
    } else {
        $identity = $IdentityServerUrl -split $OrchestratorUrl | Select-Object -last 1
        $identitySite = Get-Item IIS:\Sites\$identity
        $identityPath = $identitySite.PhysicalPath
    }
    
    $orchestratorPath = $orchestratorInstallation.Site.PhysicalPath
    $orchestratorConfigFilePath = Join-Path $orchestratorPath "UiPath.Orchestrator.dll.config"

    Write-WarningMessage "`nUpdating Orchestrator Settings"

    $orchestratorAppConfig = [xml](Get-Content $orchestratorConfigFilePath)    
    ForEach ($add in $orchestratorAppConfig.configuration.appSettings.add) {        
        Switch ($add.key) {
         "IdentityServer.Integration.Authority" { $add.value = $IdentityServerUrl }
         "ExternalAuth.System.OpenIdConnect.Authority" { $add.value = $IdentityServerUrl }
         "ExternalAuth.System.OpenIdConnect.PostLogoutRedirectUri" { $add.value = $OrchestratorUrl }
         "ExternalAuth.System.OpenIdConnect.RedirectUri" { 
             $signinsystemopenidconnect = $add.value 
             $add.value = -join ($OrchestratorUrl, "/signinsystemopenidconnect")
         }
        }
    }

    $orchestratorAppConfig.Save($orchestratorConfigFilePath)

    Write-SuccessMessage "`nOrchestrator setting got updated"

    Write-WarningMessage "`nUpdating IdentityServer AppSettings"

    $identityConfigPath = -join ($identityPath, "\appsettings.Production.json")
    $identityConfig = Get-Content $identityConfigPath    | ConvertFrom-Json
    $identityConfig.AppSettings.IdentityServerAddress = $IdentityServerUrl
    $identityConfig.AppSettings.OrchestratorUrl = $OrchestratorUrl
    $identityConfig | ConvertTo-Json -Depth 5 | set-content $identityConfigPath
    
    Write-SuccessMessage "`nIdentityServer setting got updated"
    
    Write-WarningMessage "`nUpdating ClientRedirectUris in the SqlDB"
    
    if (-not $SqlConnectionString) {
        $sqlConnectionString = $identityConfig.ConnectionStrings.DefaultConnection
    }

    $clientRedirectUri = -join ($OrchestratorUrl, "/signinsystemopenidconnect")
    
    $identityDataBaseSql = "
    UPDATE [identity].[ClientRedirectUris] SET [RedirectUri] = '$clientRedirectUri' WHERE [RedirectUri] = '$signinsystemopenidconnect'
    "

    Invoke-SqlCommand -ConnectionString $sqlConnectionString -Query $identityDataBaseSql
    
    Write-SuccessMessage "`nClientRedirectUris got updated"
    
    Write-WarningMessage "`nRestart the Orchestrator Services by iisreset"
    
    invoke-command -scriptblock { iisreset }
}

<#
    .SYNOPSIS
    Update UiPath Orchestrator SSL certificate or Identity token signing certificate.

    .DESCRIPTION
    Update the Identity token signing certificate in appsettings.Production.json.
    Update the Orchestrator SSL certificate in IIS SSL bindings.
    Run Readiness command to validate the new certificates.

    .PARAMETER SiteName
    The name of the Orchestrator site that can be found in IIS on the target machine. Usually this can be found as 'UiPath Orchestrator'.

    .PARAMETER NewTokenSigningThumbprint
    The computer certificate store thumbprint of the new token signing certificate.
    If it is set it will update appsettings.Production.json(Identity Server config) with the new certificate thumbprint.

    .PARAMETER NewTokenSigningThumbprint
    The computer certificate store thumbprint of the new Orchestrator SSL certificate
    If it is set it will update the Orchestrator IIS ssl bindings with the new certificate thumbprint.

    .PARAMETER $KeepOldCertificate
    Flag indicating if current certificate should continue to be used as a validation key for tokens that were signed with it.
#>

function Update-UiPathCertificate {
    [CmdletBinding()]
    Param(
        [Parameter(Mandatory = $true)]
        [string] $SiteName,

        [Parameter(Mandatory = $false)]
        [string] $NewTokenSigningThumbprint,

        [Parameter(Mandatory = $false)]
        [string] $NewSSLThumbprint,

        [Parameter(Mandatory = $false)]
        [bool] $KeepOldCertificate = $true
    )
    
    $orchestratorInstallation = Get-OrchestratorInstallation -SiteName $SiteName
    if (-not $orchestratorInstallation) {
        Write-ErrorMessage "Orchestrator is not installed. Can not update UiPath token signing and Orchestrator SSL certificates."
        return
    }
    
    if ($NewTokenSigningThumbprint) {
        $identitySite = Get-IdentitySite -OrchestratorSiteName $orchestratorInstallation.Site.Name
        if (-not $identitySite) {
            Write-ErrorMessage "Identity is not installed. Can not update token signing certificate."
            return
        }

        $identityPath = $identitySite.PhysicalPath
        $identityConfigPath = -join ($identityPath, "\appsettings.Production.json")
        $identityConfig = Get-Content $identityConfigPath    | ConvertFrom-Json

        Write-WarningMessage "`nUpdating Certificate Details in IdentityServer App Settings"

        if ($KeepOldCertificate)
        {
            $signingCredential=$config.AppSettings.SigningCredentialSettings.StoreLocation
            $validationKeys = $config.AppSettings.SigningCredentialSettings.StoreLocation.ValidationKeys
            if ($validationKeys -ne $null)
            {
                $validationKeys += New-Object -TypeName PSObject -Property @{Name=$signingCredential.Name; Location=$signingCredential.Location; NameType=$signingCredential.NameType}
                $config.AppSettings.SigningCredentialSettings.StoreLocation.ValidationKeys = $validationKeys
            }
            else
            {
                $validationKeys=@()
                $validationKeys += New-Object -TypeName PSObject -Property @{Name=$signingCredential.Name; Location=$signingCredential.Location; NameType=$signingCredential.NameType}
                $config.AppSettings.SigningCredentialSettings.StoreLocation | Add-Member -Name "ValidationKeys" -Value $validationKeys -MemberType NoteProperty -Force
            }
        }

        $identityConfig.AppSettings.SigningCredentialSettings.StoreLocation.Name = $NewTokenSigningThumbprint

        $identityConfig | ConvertTo-Json -Depth 5 | set-content $identityConfigPath

        Restart-WebAppPool $identitySite.applicationPool

        Write-SuccessMessage "`nIdentityServer App Setting got updated"
    }
        
    if($NewSSLThumbprint){
        Write-WarningMessage "`nUpdating IIS binding with new SSL Certificate"

        $binding = (Get-ChildItem -Path IIS:\SSLBindings | Where-Object Sites -eq $SiteName)[0]
        $newSSLThumbprintPath = "cert:\LocalMachine\$($binding.Store)\$($NewSSLThumbprint)"
        Remove-Item -path IIS:\SslBindings\$($binding.IPAddress.IPAddressToString)!$($binding.Port)
        Get-Item -Path $newSSLThumbprintPath | New-item -path IIS:\SslBindings\$($binding.IPAddress.IPAddressToString)!$($binding.Port)

        Restart-WebAppPool $SiteName 

        Write-SuccessMessage "`nNew SSL Certificate got updated"
    }
    
    Write-WarningMessage "`nRestart the Orchestrator Services by iisreset"
    
    invoke-command -scriptblock { iisreset }
}