Using namespace System.Security.AccessControl;
Using namespace System.Security.Cryptography.X509Certificates;

function Test-PlatformCertificates {
    param (
        [Parameter(Mandatory = $true)]
        $IdentitySite,

        [Parameter(Mandatory = $true)]
        $IdentityConfig,

        [Parameter(Mandatory = $true)]
        [string]$OrchestratorSiteName
    )

    Write-Host "Checking platform certificates..."
    $validationResult = [ValidationResult]::new()

    if (-not $IdentitySite -or -not $IdentityConfig) {
        Write-ErrorMessage "Identity is not installed. Can not complete the Identity pre-installation checks."
        $validationResult.Errors++
        return $validationResult
    }

    if ($IdentityConfig.ConfigEncryptionSettings) {
        Write-ErrorMessage "appsettings.production.json is encrypted."
        
        $validationResult.Errors++
        return $validationResult
    }
    
    $certificate = Get-IdentityCertificate -IdentitySite $IdentitySite -IdentityConfig $IdentityConfig
    
    if (-not $certificate) {
        $validationResult.Errors++
        return $validationResult
    }

    $validationResult += Test-IdentityCertificatePublicKey -IdentitySite $IdentitySite -Certificate $certificate

    $validationResult += Test-IdentityCertificatePrivateKey -IdentitySite $IdentitySite -Certificate $certificate

    Set-ClippedValidationResult -ValidationResult $validationResult
    Set-SuccessesEmptyValidationResult -ValidationResult $validationResult

    Write-MessageWithValidation -Message "Platform certificates validation is finished." -ValidationResult $ValidationResult

    return $validationResult
}

function Test-IdentityCertificatePublicKey {

    param(
        [Parameter(Mandatory = $true)]
        $IdentitySite,

        [Parameter(Mandatory = $true)]
        $Certificate
    )
    
    Write-Verbose "Checking Identity token signing certificate public key..."
    
    $certificateKeySize = $certificate.PublicKey.Key.KeySize

    if ($certificateKeySize -lt 2048) {
        Write-ErrorMessage "Key size must be at least 2048 but was $certificateKeySize."
        $validationResult.Errors++
    }  else {
        Write-Verbose "Identity token signing certificate public key size is $certificateKeySize."
        $validationResult.Successes++
    }

    return $validationResult
}

function Test-IdentityCertificatePrivateKey {
    param(
        [Parameter(Mandatory = $true)]
        $IdentitySite,

        [Parameter(Mandatory = $true)]
        $Certificate
    )

    Write-Verbose "Checking Identity token signing certificate private key..."
    $validationResult = [ValidationResult]::new()
    
    $appPoolName = $identitySite.applicationPool
    $appPool = Get-Item IIS:\AppPools\$appPoolName
    
    $appPoolUser = if ($appPool.processModel.identityType -eq "ApplicationPoolIdentity") {
        $appPool.name
    }else {
        $appPool.processModel.userName
    }

    if (-not $certificate.HasPrivateKey) {
        Write-ErrorMessage "Identity token signing certificate does not have a private key."
        
        $validationResult.Errors++
        return $validationResult
    }

    $certPkAccess = Get-CertPrivateKeyACL $Certificate
    if (-not $certPkAccess)
    {
        Write-WarningMessage "Could not determine access rights for the private key of the signing certificate. For help, see https://docs.uipath.com/installation-and-upgrade/docs/using-a-certificate-for-the-https-protocol#internal-server-error"
        $validationResult.Warnings++
        return $validationResult
    }

    $userAccess = $certPkAccess | Where-Object { $_.IdentityReference.Value -like "*$appPoolUser" }
    
    if (-not $userAccess) {
        $usersWithAccess = $certPkAccess.IdentityReference.Value -join ', '

        Write-ErrorMessage "AppPool user $appPoolUser does not have access to the private key certificate. Users with access are $usersWithAccess. For help, see https://docs.uipath.com/orchestrator/docs/identity-server-troubleshooting#section-keyset-does-not-exist-error-after-installation"
        $validationResult.Errors++
    } elseif ($userAccess.AccessControlType -eq "Deny") {
        Write-ErrorMessage "AppPool user $appPoolUser is denied access to the private key certificate. For help, see https://docs.uipath.com/orchestrator/docs/identity-server-troubleshooting#section-keyset-does-not-exist-error-after-installation"
        $validationResult.Errors++
    } elseif ( ($null -ne $userAccess.CryptoKeyRights) -and
        (-not (
            $userAccess.CryptoKeyRights.HasFlag([CryptoKeyRights]::GenericRead) -or
            $userAccess.CryptoKeyRights.HasFlag([CryptoKeyRights]::ReadData)))) {
        Write-ErrorMessage "AppPool user $appPoolUser is missing rights to the certificate private key. For help, see https://docs.uipath.com/orchestrator/docs/identity-server-troubleshooting#section-keyset-does-not-exist-error-after-installation"
        $validationResult.Errors++
    } else {
        Write-Verbose "Identity token signing certificate private key has proper access rights."
        $validationResult.Successes++
    }

    return $validationResult
}

function Get-IdentityCertificate {
    param (
        [Parameter(Mandatory = $true)]
        $IdentitySite,
        
        [Parameter(Mandatory = $true)]
        $IdentityConfig
    )

    $identityCertificateLocation = $IdentityConfig.AppSettings.SigningCredentialSettings.StoreLocation.Location
    if ([string]::IsNullOrWhiteSpace($identityCertificateLocation)) {
        Write-ErrorMessage "Identity token signing certificate location is not set in appsettings.Production.json."
        return $null
    }

    $identityCertificateName = $IdentityConfig.AppSettings.SigningCredentialSettings.StoreLocation.Name
    if ([string]::IsNullOrWhiteSpace($identityCertificateName)) {
        Write-ErrorMessage "Identity token signing certificate name is not set in appsettings.Production.json."
        return $null
    }

    $certificatePath = "Cert:\$identityCertificateLocation\My\$identityCertificateName"
    $certificate = Get-Item $certificatePath -ErrorAction SilentlyContinue

    if (-not $certificate) {
        Write-ErrorMessage "Identity token signing certificate could not be found at '$certificatePath'."
    } else {
        Write-Verbose "Identity token signing certificate found at '$certificatePath'."
    }

    return $certificate
}

function Get-CertPrivateKeyACL 
{
    param (
        [Parameter(Mandatory = $true)]
        $certificate
    )

    # get ACL for private keys exportable
    if ($null -ne $certificate.PrivateKey)
    {
        $access = $certificate.PrivateKey.CspKeyContainerInfo.CryptoKeySecurity.Access
    } 
    else {
        # in .net 4.6.2 Private key access is changed and only private key with export rights can be investigated
        # https://docs.microsoft.com/en-us/dotnet/framework/whats-new/#Crypto462
        if ("RSA" -eq $certificate.PublicKey.Oid.FriendlyName )
        {
            $rsaCert = [System.Security.Cryptography.X509Certificates.RSACertificateExtensions]::GetRSAPrivateKey($certificate)
            $uniqueKey =$rsaCert.key.UniqueName
            $path = "$env:ALLUSERSPROFILE\Microsoft\Crypto\Keys\$uniqueKey"
            $permissions = Get-Acl -Path $path
            $access = $permissions.Access
        }
    }
    return $access
}