function Test-WebCoreHostModuleInstallation {
    Write-Host "Checking AspNetCore hosting module..."
    $validationResult = [ValidationResult]::new()

    $aspNetCoreModule = Get-WebGlobalModule AspNetCoreModuleV2
    $netCoreHostingDirectoryRegistry = "HKLM:\SOFTWARE\WOW6432Node\Microsoft\Updates\.NET Core"
    
    if (-not $aspNetCoreModule -or -not (Test-Path $netCoreHostingDirectoryRegistry)) {
        $errorMessage = "AspNetCoreModuleV2 not found. Check that the ASP.NET Core Hosting Bundle is properly installed."
        $errorMessage += [Environment]::NewLine + "https://docs.microsoft.com/en-us/aspnet/core/host-and-deploy/iis/?view=aspnetcore-3.1#install-the-net-core-hosting-bundle"
        
        Write-ErrorMessage $errorMessage
        $validationResult.Errors++
        return $validationResult
    }

    $netCoreHostingBundleRegistryVersions = Get-ChildItem -path $netCoreHostingDirectoryRegistry -Recurse | 
        Where-Object { $_.Name -like '*Microsoft .NET Core *'} |
        Get-ItemProperty -Name "PackageVersion" |
        Where-Object {[Version]::new($_.PackageVersion) -ge [version]::new(3, 1)}

    if ($netCoreHostingBundleRegistryVersions.Length -eq 0) {
        $errorMessage = "ASP.NET Core Hosting Bundle V3.1 is required. Check that the ASP.NET Core Hosting Bundle is properly installed."
        $errorMessage += [Environment]::NewLine + "https://docs.microsoft.com/en-us/aspnet/core/host-and-deploy/iis/?view=aspnetcore-3.1#install-the-net-core-hosting-bundle"
        
        Write-ErrorMessage $errorMessage
        $validationResult.Errors++
        return $validationResult
    }

    Write-SuccessMessage "AspNetCore hosting module is installed."
    $validationResult.Successes++

    return $validationResult
}

function Test-OrchestratorCyberArkAIMInstallation {
    param(
        [Parameter(Mandatory = $true)]
        [System.Xml.XmlDocument] $WebConfig
    )

    $validationResult = [ValidationResult]::new()

    $plugins = Read-WebConfigKey $WebConfig "Plugins.SecureStores"

    if (-not ($plugins -like "*UiPath.Orchestrator.CyberArk.dll*")) {
        # CyberArk plugin is not used in Orchestrator.
        return [ValidationResult]::new()
    }
    
    Write-Host "Checking CyberArk Application Access Manager installation..."

    $cyberArkAimRegistryPath = "HKLM:\SOFTWARE\WOW6432Node\CyberArk\CyberArk Application Identity Manager"
    $cyberArkAamRegistryPath = "HKLM:\SOFTWARE\WOW6432Node\CyberArk\CyberArk Application Access Manager"
    if (-not (Test-Path $cyberArkAimRegistryPath) -and -not (Test-Path $cyberArkAamRegistryPath)) {
        # The path does not exist in Windows registry. CyberArk AIM is not installed.
        Write-WarningMessage "CyberArk AAM is not installed, but the plugin 'UiPath.Orchestrator.CyberArk.dll' is set for 'Plugins.SecureStores' key in web.config."
        $validationResult.Warnings++
        return $validationResult
    }

    $validationResult = [ValidationResult]::new()

    $cyberArk = Join-Path ${Env:ProgramFiles(x86)} "CyberArk"
    $cyberArkSDK = Join-Path $cyberArk "ApplicationPasswordSdk"
    $cliPasswordSDK = Join-Path $cyberArkSDK "CLIPasswordSDK.exe"
    
    if (Test-Path $cliPasswordSDK) {
        Write-SuccessMessage "CyberArk CLIPasswordSDK.exe was found at the default installation path '$cliPasswordSDK'."
        $validationResult.Successes++
    }
    else {
        $warningMessage = "CyberArk CLIPasswordSDK.exe was not found at the default installation path '$cliPasswordSDK'."
        $warningMessage += "If you are using CyberArk credential store "
        $warningMessage += "please add in Web.Config appSettings section the following Plugins.SecureStores.CyberArk.CLIPasswordSDKExePath key with the value to the CyberArk CLIPasswordSDK.exe path."

        Write-WarningMessage $warningMessage
        $validationResult.Warnings++
    }
    return $validationResult
}

function Test-OrchestratorWebConfigEncryption {
    param(
        [Parameter(Mandatory = $true)]
        [System.Xml.XmlDocument] $WebConfig
    )

    Write-Host "Checking Web.config sections encryption..."
    $validationResult = [ValidationResult]::new()

    $encryptedSections = $WebConfig.SelectNodes("/configuration/*[@configProtectionProvider]")
    if ($encryptedSections.count -gt 0) {
        $errorMessage = "Web.config is encrypted. Before upgrading to 2020.10 please decrypt the following protected sections: {0}." -f ($encryptedSections -join ",")
        $errorMessage += [Environment]::NewLine + "https://docs.uipath.com/orchestrator/docs/encrypting-webconfig-sections"

        Write-ErrorMessage $errorMessage
        $validationResult.Errors++
    } else {
        Write-SuccessMessage "Web.config sections are not encrypted."
        $validationResult.Successes++
    }

    return $validationResult
}

function Test-OrchestratorWebConfigLockedSections {
    
    Write-Host "Checking IIS configuration locked sections..."
    $validationResult = [ValidationResult]::new()

    $sections = @("modules","handlers","aspNetCore","security","httpProtocol","httpErrors","rewrite","staticContent","httpCompression","defaultDocument","directoryBrowse")
    $lockedSections = @()

    foreach ($section in $sections) {
        $sectionLock = Get-WebConfigurationLock -Filter "//$section" -PSPath "IIS:\"    
        if ($sectionLock.Value -eq "Deny") {
            $lockedSections += $section
        }
    }

    if ($lockedSections.Length -ne 0) {
        $errorMessage = "Some configuration sections are locked in IIS. Before upgrading to 2020.10 please unlock the following locked sections from 'system.WebServer': {0}." -f ($lockedSections -join ",")
        $errorMessage += [Environment]::NewLine + "https://docs.microsoft.com/en-us/iis/get-started/planning-for-security/how-to-use-locking-in-iis-configuration"

        Write-ErrorMessage $errorMessage
        $validationResult.Errors++
    } else {
        Write-SuccessMessage "Configuration sections are not locked."
        $validationResult.Successes++
    }

    return $validationResult
}

function Test-OrchestratorCertificateSubject {
    param(
        [Parameter(Mandatory = $false)]
        $IdentityConfig,

        [Parameter(Mandatory = $true)]
        [string]$OrchestratorSiteName
    )

    Write-Host "Checking Orchestrator ssl certificate subject alternative names..."
    $validationResult = [ValidationResult]::new()

    $orchestratorHost = Get-OrchestratorHostName -IdentityConfig $IdentityConfig -OrchestratorSiteName $OrchestratorSiteName

    if ([string]::IsNullOrWhiteSpace($orchestratorHost)) {
        Write-ErrorMessage "Orchestrator host name could not be determined."
        $validationResult.Errors++
        return $validationResult
    }

    $sslCertificate = Get-OrchestratorCertificate -OrchestratorSiteName $OrchestratorSiteName
    if (-not $sslCertificate) {
        $validationResult.Errors++
        return $validationResult
    }

    if ($sslCertificate.DnsNameList) {
        $dnsNameList = $sslCertificate.DnsNameList
    } else {
        # DnsNameList property does not exist in the ssl certificate.
        # Lookup subject alternative names in the certificate extensions property.
        $sanExtension = ($sslCertificate.Extensions | Where-Object {$_.Oid.FriendlyName -eq "subject alternative name"}).Format(1)
        $dnsNameList = $sanExtension.Split([Environment]::NewLine) |
            Where-Object {$_ -like "DNS Name=*"} |
            Select-Object @{ Name = "DNSName"; Expression = {($_ -split "DNS Name=")}} |
            Select-Object -ExpandProperty DNSName
    }

    $orchestratorHostDNS = $dnsNameList | Where-Object {$orchestratorHost -like $_}
    if ($orchestratorHostDNS) {
        Write-SuccessMessage "Orchestrator host name is valid for the ssl certificate subject alternative names."
        $validationResult.Successes++
    } else {
        Write-ErrorMessage "Orchestrator host $orchestratorHost is not a subject alternative name of the SSL certificate. Subject alternative names are $dnsNameList."
        $validationResult.Errors++
    }
    return $validationResult
}

function Test-OrchestratorSqlServerScaleoutUse {
    param(
        [Parameter(Mandatory = $true)]
        [System.Xml.XmlDocument] $WebConfig
    )

    Write-Host "Checking sql server scaleout use..."

    $validationResult = [ValidationResult]::new()

    if (Confirm-EncryptedSection -WebConfig $WebConfig -Section "appSettings") {
        Write-WarningMessage "Web.config section 'appSettings' is encrypted, could not determine the sql server scaleout use."
        $validationResult.Warnings++
        return $validationResult
    }

    $sqlServerScaleout = Read-WebConfigKey -WebConfig $WebConfig -Key "LoadBalancer.UseSqlServer"
    if ($sqlServerScaleout -eq "true") {
        Write-WarningMessage "After 20.10 installation the LoadBalancer.UseSqlServer key from Web.config will be migrated to LoadBalancer.UseRedis key in UiPath.Orchestrator.dll.config"
        $validationResult.Warnings++
    } else {
        Write-SuccessMessage "Sql server scaleout is not used."
        $validationResult.Successes++
    }

    return $validationResult
}

function Test-OrchestratorCredentialStoresPlugins {
    param(
        [Parameter(Mandatory = $true)]
        [System.Xml.XmlDocument] $WebConfig,
        [Parameter(Mandatory = $true)]
        [string] $OrchestratorDirectory
    )
    Write-Host "Checking external credential store plugins target framework..."
    $validationResult = [ValidationResult]::new()
    
    if (Confirm-EncryptedSection -WebConfig $WebConfig -Section "appSettings") {
        Write-WarningMessage "Web.config section 'appSettings' is encrypted, could not find the credential store plugins."  
        $validationResult.Warnings++
        return $validationResult
    }

    $credentialStorePluginsPath = Get-CredentialStoresPluginsPath -WebConfig $WebConfig -OrchestratorDirectory $OrchestratorDirectory
    if (Test-Path $credentialStorePluginsPath) {
        Write-Verbose "Credential store plugins directory found at $credentialStorePluginsPath."
    }
    else {
        Write-WarningMessage "Credential store plugins directory not found at $credentialStorePluginsPath."
        
        $validationResult.Warnings++
        return $validationResult
    }

    $pluginNames = Read-WebConfigKey -WebConfig $WebConfig -Key "Plugins.SecureStores"

    if([string]::IsNullOrWhiteSpace($pluginNames)){
        Write-Verbose "No credential stores plugins to validate."
        $validationResult.Successes++
    }
    else {

        $plugins = $pluginNames.split(",", [StringSplitOptions]::RemoveEmptyEntries) |
            Get-Unique |
            Select-Object @{ Name="Assembly"; Expression = {$_.Trim()}} |
            Join-Path -Path $credentialStorePluginsPath -ChildPath {$_.Assembly}
    
        $excludedCredentialStorePlugins = @("UiPath.Orchestrator.CyberArk.dll", "UiPath.Orchestrator.AzureKeyVault.SecureStore.dll")

        $validationResult = Test-PluginsTargetFramework -Plugins $plugins -ExcludedPlugins $excludedCredentialStorePlugins
    }

    Write-MessageWithValidation `
        -Message "Credential stores plugins validation is finished." `
        -ValidationResult $validationResult
    return $validationResult
}

function Test-OrchestratorNLogPlugins {
    param(
        [Parameter(Mandatory = $true)]
        [System.Xml.XmlDocument] $WebConfig,
        [Parameter(Mandatory = $true)]
        [string] $OrchestratorBinPath
    )
    Write-Host "Checking external NLog plugins target framework..."
    $validationResult = [ValidationResult]::new()

    $nLogPluginsPath = $OrchestratorBinPath
    if (Test-Path $nLogPluginsPath) {
        Write-Verbose "Searching for NLog plugins in '$nLogPluginsPath' directory."
    }
    else {
        Write-ErrorMessage "NLog plugins directory was not found at '$nLogPluginsPath'."
        
        $validationResult.Errors++
        return $validationResult
    }

    $namespaceManager = New-Object System.Xml.XmlNamespaceManager $webConfig.NameTable
    $namespaceManager.AddNameSpace("ns", "http://www.nlog-project.org/schemas/NLog.xsd")

    if (Confirm-EncryptedSection -WebConfig $WebConfig -Section "ns:nlog" -namespaceManager $namespaceManager) {
        Write-WarningMessage "Web.config section 'nlog' is encrypted, could not find the NLog plugins."
        $validationResult.Warnings++
        return $validationResult
    }

    $nLogPluginsKeys = $WebConfig.SelectNodes("//ns:nlog/ns:extensions/ns:add", $namespaceManager)

    if([string]::IsNullOrWhiteSpace($nLogPluginsKeys)){
        Write-Verbose "No NLog plugins to validate."
        $validationResult.Successes++
    }
    else {
        $nLogPluginNames = $nLogPluginsKeys |
            Select-Object @{ Name="Assembly"; Expression = {$_.Attributes["assembly"].Value.Trim() + ".dll"} } |
            Join-Path -Path $nLogPluginsPath -ChildPath {$_.Assembly}

        $excludedNLogPlugins = @("NLog.Targets.ElasticSearch.dll", "NLog.Mongo.dll", "Microsoft.ApplicationInsights.NLogTarget.dll")
        $validationResult = Test-PluginsTargetFramework -Plugins $nLogPluginNames -ExcludedPlugins $excludedNLogPlugins
    }

    Write-MessageWithValidation `
        -Message "NLog plugins validation is finished." `
        -ValidationResult $validationResult
    return $validationResult
}

function Test-PluginsTargetFramework {
    param(
        [Parameter(Mandatory = $true)]
        [string[]] $Plugins,
        [Parameter(Mandatory = $true)]
        [string[]] $ExcludedPlugins
    )
    $validationResult = [ValidationResult]::new()

    foreach ($plugin in $Plugins) {
        if ([string]::IsNullOrWhiteSpace($plugin)) {
            continue
        }
        $isPluginExcluded = ($ExcludedPlugins | Where-Object {$plugin.EndsWith($_)}).Count -gt 0
        if ($isPluginExcluded) {
            Write-Verbose "Skipping plugin validation for '$plugin'."
            continue
        }

        Write-Verbose "Checking plugin '$plugin' ..."

        if (-not (Test-Path $plugin)) {
            Write-ErrorMessage "Plugin '$plugin' does not exist."
            $validationResult.Errors++
            continue
        }

        $targetFramework = [Reflection.Assembly]::LoadFrom($plugin).CustomAttributes |
            Where-Object {$_.AttributeType.Name -eq "TargetFrameworkAttribute"} | 
            Select-Object -ExpandProperty ConstructorArguments | 
            Select-Object -ExpandProperty Value
        
        if (-not ($targetFramework)) {
            Write-WarningMessage "Target framework could not be determined for '$plugin'."
            $validationResult.Warnings++
            continue
        }

        $items = $targetFramework.split(",")
        $frameworkName = $items[0]

        if ($frameworkName -like "*NetFramework*") {
            Write-ErrorMessage "Target NetStandard 2.0 for plugin '$plugin'."
            $validationResult.Errors++
            continue
        }

        $frameworkVersionLabels = $items |
            Select-Object -Skip 1 |
            Where-Object {$_.StartsWith("Version=v", [StringComparison]::OrdinalIgnoreCase)} |
            Where-Object {$_.length -gt 9}

        if (-not ($frameworkVersionLabels)) {
            Write-WarningMessage "Target framework version could not be found for '$plugin'."
            $validationResult.Warnings++
            continue
        }
        
        $frameworkVersion = [System.Version]::new()
        foreach ($frameworkVersionLabel in $frameworkVersionLabels) {
            if ([System.Version]::TryParse($frameworkVersionLabel.Substring(9), [ref] $frameworkVersion)) {
                break
            }
        }

        if (-not ($frameworkVersion)) {
            Write-WarningMessage "Target framework version could not be found for '$plugin'."
            $validationResult.Warnings++
            continue
        }

        Write-Verbose "Plugin '$plugin' validation succeeded."
        $validationResult.Successes++
    }

    Set-SuccessesEmptyValidationResult -ValidationResult $validationResult

    Write-ValidationResult $validationResult -ValidationSubject "Plugin" -OutputVerbose $true

    Set-ClippedValidationResult -ValidationResult $validationResult

    return $validationResult
}

function Test-OrchestratorFileSystemBuckets {
    param(
        [Parameter(Mandatory = $true)]
        [System.Xml.XmlDocument] $AppConfig
    )
    Write-Host "Checking buckets with file system storage provider..."
    $validationResult = [ValidationResult]::new()

    if (Confirm-EncryptedSection -WebConfig $AppConfig -Section "connectionStrings") {
        Write-WarningMessage "Could not determine if any buckets with file system provider are in use. Config section 'connectionString' is encrypted, could not find the sql connection string to the UiPath database."  
        $validationResult.Warnings++
        return $validationResult
    }

    if (Confirm-EncryptedSection -WebConfig $AppConfig -Section "appSettings") {
        Write-WarningMessage "Could not determine if any buckets with file system provider are in use. Config section 'appSettings' is encrypted, could not find the allowed list paths."
        $validationResult.Warnings++
        return $validationResult
    }

    $allowedPathsValue = Read-WebConfigKey -WebConfig $AppConfig -Key "Buckets.FileSystem.Allowlist"
    if (-not [string]::IsNullOrWhiteSpace($allowedPathsValue)){
        $allowedListPaths = $allowedPathsValue.Split('|') |
            Where-Object {-not [string]::IsNullOrWhiteSpace($_)} |
            Select-Object @{Name ="Path"; Expression = { $_.Trim()}}
    } elseif ($null -ne $allowedPathsValue -and $allowedPathsValue.Trim() -eq [string]::Empty) {
        # Config exists but it is empty.
        $allowedListPaths = @()
    }

    $validationResult += Test-FullyQualifiedPaths -Paths $allowedListPaths -Source "the config file allowed list section"

    $defaultConnectionStringNode = $AppConfig.SelectSingleNode("/configuration/connectionStrings/add[@name='Default']")
    if (-not $defaultConnectionStringNode -or -not $defaultConnectionStringNode.Attributes["connectionString"]) {
        Write-WarningMessage "Could not determine if any buckets with file system provider are in use. Default UiPath connection string was not found."
        $validationResult.Warnings++

        Set-ClippedValidationResult -ValidationResult $validationResult
        return $validationResult
    }

    $uipathDbConnectionString = $defaultConnectionStringNode.Attributes["connectionString"].Value
    $filesSystemBucketsQuery = "
        SELECT [Name], [StorageParameters]
        FROM [dbo].[Buckets]
        WHERE [IsDeleted] = 0 and [StorageProvider] = 'FileSystem'"
    
    try{
        $databaseData = Invoke-SqlCommand -ConnectionString $uipathDbConnectionString -Query $filesSystemBucketsQuery
    } catch {
        Write-WarningMessage "Could not determine if any buckets with file system provider are in use. Could not connect to the UiPath Database."
        $validationResult.Warnings++

        Set-ClippedValidationResult -ValidationResult $validationResult
        return $validationResult
    }

    $skipRootPathPrefixLength = "RootPath=".Length;
    $buckets = $databaseData |
        Select-Object @{ Name="Path"; Expression = {$_.StorageParameters.Substring($skipRootPathPrefixLength).Trim()}},
            @{ Name = "Name"; Expression = {$_.Name}}

    $validationResult += Test-FullyQualifiedPaths -Paths $buckets -Source "the buckets with file system provider"

    $allowedListPaths = $allowedListPaths | Select-Object -ExpandProperty Path

    $databasePaths = $buckets | Select-Object -ExpandProperty Path
    $databaseBasePaths = Get-BasePaths -Paths $databasePaths

    $anyUnqualifiedPath = $validationResult.Warnings -gt 0
    $validationResult += Test-AllowedListPaths $databaseBasePaths $allowedListPaths $anyUnqualifiedPath

    Set-ClippedValidationResult -ValidationResult $validationResult
    return $validationResult
}

function Test-AllowedListPaths {
    param(
        [string[]] $DatabasePaths,
        [string[]] $allowedListPaths,
        [bool] $AnyUnqualifiedPath
    )
    $validationResult = [ValidationResult]::new()

    if ($DatabasePaths.Count -gt 0 ) {
        # There are buckets using file system provider
        [string[]]$unlistedPaths = @()
        if (-not $allowedListPaths) {
            # Where-object will iterate over empty collection array properties itself
            $unlistedPaths = $DatabasePaths
        } else {
            foreach ($databasePath in $DatabasePaths) {
                [array]$baseDirPaths = $allowedListPaths | Where-Object {Test-SubPathOf -Path $databasePath -BaseDirPath $_}
                if ($baseDirPaths.Count -eq 0) {
                    # The database bucket path is not on the allowed list if it's not a subpathof of any allowed list paths
                    $unlistedPaths += $databasePath
                }
            }
        }
        $unlistedPathsText = $unlistedPaths -join '|'
        if (-not $allowedListPaths) {
            # Buckets.FileSystem.Allowlist is not set
            Write-ErrorMessage "All storage buckets using the file system provider are not on the allowed list. "
            Write-WarningMessage "Add the following setting in the configuration file to allow all exiting buckets root paths:"
            Write-WarningMessage "      <add key=""Buckets.FileSystem.Allowlist"" value=""$unlistedPathsText"" />      "

            $validationResult.Errors++
        } elseif ($unlistedPaths.Count -gt 0) {
            # Buckets.FileSystem.Allowlist is set but might not contain all buckets root paths
            Write-WarningMessage "There are some storage buckets using the file system provider that are not on the allowed list. The buckets feature will not work for buckets with root paths that are not on the allowed list."
            Write-WarningMessage "Check if any of the following paths are required to be on the allowed list and add them to the 'Buckets.FileSystem.Allowlist' key in configuration file:"
            Write-WarningMessage "|$unlistedPathsText"

            $validationResult.Warnings++
        } elseif (-not $AnyUnqualifiedPath){
            # All Buckets files system provider root paths are on the allowed list
            Write-SuccessMessage "All storage buckets using file system provider have the root path on the allowed list in the configuration file." 
            $validationResult.Successes++
        }
    } elseif (-not $AnyUnqualifiedPath) {
        # There is not bucket using file system provider
        Write-SuccessMessage "There are no buckets with file system provider in use." 
        $validationResult.Successes++
    }
    return $validationResult
}

function Test-FullyQualifiedPaths {
    param (
        $Paths,
        [string]
        $Source
    )
    $validationResult = [ValidationResult]::new()
    if (-not $Paths) {
        return $validationResult
    }

    [array]$unqualifiedPaths = $Paths | Where-Object {-not (Test-FullyQualifiedPath -Path ($_.Path))}
    if ($unqualifiedPaths.Count -gt 0) {
        Write-WarningMessage "Some paths from $Source are invalid. Only fully qualified paths are allowed."
        Write-WarningMessage "Fully qualify the following paths from $Source :"

        foreach ($path in $unqualifiedPaths) {
            if ($path.Name) {
                Write-WarningMessage "Name: $($path.Name), Path: $($path.Path)"
            } else {
                Write-WarningMessage "Path: $($path.Path)"
            }
        }
        $validationResult.Warnings++
    }
    else{
        Write-Verbose "All paths from $Source are fully qualified."
        $validationResult.Successes++
    }

    return $validationResult
}

function Test-FullyQualifiedPath {
    param (
        [Parameter(Mandatory = $true)]
        [string]
        $Path
    )
    if ([string]::IsNullOrWhiteSpace($Path)) {
        return $false
    }

    if ($Path.IndexOfAny([System.IO.Path]::GetInvalidPathChars()) -ne -1 `
        -or -not [System.IO.Path]::IsPathRooted($Path)) {
        return $false
    }

    $pathRoot = [System.IO.Path]::GetPathRoot($Path)
    if ($pathRoot.Length -le 2) {
        return $false
    }
    if ($pathRoot[0] -ne '\' -or $pathRoot[1] -ne '\') {
        return $true
    }
    return $pathRoot.Trim('\').IndexOf('\') -ne -1
}

function Get-BasePaths {
    param(
        [string[]] $Paths
    )
    $basePaths = @()
    if (-not $Paths) {
        return $basePaths
    }
    # Get unique paths. 'C:\Home' and 'C:Home\' should be considered as the same path
    $Paths = $Paths | Where-Object {(Test-FullyQualifiedPath -Path $_)} |
        Select-Object -Unique @{Name = "Path"; Expression = {Get-NormalizedPath -Path $_}} |
        Select-Object -ExpandProperty Path
    
    foreach ($path in $Paths) {
        $subPaths = $Paths | Where-Object {Test-SubPathOf -Path $path -BaseDirPath ($_)}
        if ($subPaths.Count -eq 1) {
            # If the tested path is a sub path of itself and it is a valid path
            $basePaths += $path
        }
    }

    return $basePaths
}

function Test-SubPathOf {
    param(
        [Parameter(Mandatory = $true)]
        [string]
        $Path,
        [Parameter(Mandatory = $true)]
        [string]
        $BaseDirPath
    )
    if (-not (Test-FullyQualifiedPath -Path $Path) -or -not (Test-FullyQualifiedPath -Path $BaseDirPath)) {
        return $false
    }

    $Path = Get-NormalizedPath -Path $Path
    $BaseDirPath = Get-NormalizedPath -Path $BaseDirPath

    return $Path.StartsWith($BaseDirPath)
}

function Get-NormalizedPath {
    param(
        [Parameter(Mandatory = $true)]
        [string]
        $Path
    )
    $Path = $Path.Replace('/', [System.IO.Path]::DirectorySeparatorChar).Replace('\', [System.IO.Path]::DirectorySeparatorChar)

    if (-not ($Path.EndsWith([System.IO.Path]::DirectorySeparatorChar, [StringComparison]::InvariantCulture))) {
        $Path += [System.IO.Path]::DirectorySeparatorChar
    }
    return [System.IO.Path]::GetFullPath($Path)
}

function Get-CredentialStoresPluginsPath {
    param(
        [Parameter(Mandatory = $true)]
        [System.Xml.XmlDocument] $WebConfig,
        [Parameter(Mandatory = $true)]
        [string] $OrchestratorDirectory
    )
    $pluginsFolder = Read-WebConfigKey -WebConfig $webConfig -Key "Plugins.SecureStores.Path"
    if (-not $pluginsFolder) {
        $pluginsFolder = "Plugins"
    }

    $pluginsFolderPath = Join-Path $OrchestratorDirectory $pluginsFolder
    return $pluginsFolderPath
}

function Get-OrchestratorCertificate {
    param (
        [Parameter(Mandatory = $true)]
        [string]$OrchestratorSiteName
    )

    $sslBindings = Get-ChildItem -Path IIS:\SSLBindings | Where-Object Sites -eq $OrchestratorSiteName
    if ($sslBindings.Count -eq 0) {
        Write-ErrorMessage "No SSL binding is used by $OrchestratorSiteName."
        return $null
    }

    $sslBinding = $sslBindings[0]
    $sslCertificatePath = "Cert:\LocalMachine\$($sslBinding.Store)\$($sslBinding.Thumbprint)"
    $sslCertificate = Get-Item $sslCertificatePath -ErrorAction SilentlyContinue
    if (-not $sslCertificate) {
        Write-ErrorMessage "Orchestrator ssl certificate could not be found at '$sslCertificatePath'."
    } else {
        Write-Verbose "Orchestrator ssl certificate found at '$sslCertificatePath'."
    }

    return $sslCertificate
}

function Get-OrchestratorHostName {
    param(
        [Parameter(Mandatory = $false)]
        $IdentityConfig,

        [Parameter(Mandatory = $true)]
        [string]$OrchestratorSiteName
    )

    if ($IdentityConfig) {
        [URI]$orchestratorUri = $IdentityConfig.AppSettings.OrchestratorUrl
        $orchestratorHost = $orchestratorUri.DnsSafeHost
    } else {
        $orchestratorHost = [System.Net.Dns]::GetHostEntry([string]$env:computername).HostName
    }
    $binding = Get-WebSite "$OrchestratorSiteName" | Get-WebBinding | Select-Object -ExpandProperty BindingInformation
    $bindingOrchestratorHost = ($binding -split ":")[-1]
    
    if ([string]::IsNullOrWhiteSpace($bindingOrchestratorHost)) {
        return $orchestratorHost
    } elseif ($orchestratorHost -ne $bindingOrchestratorHost) {
        Write-ErrorMessage "Orchestrator IIS binding host name and Identity Server Orchestrator Url are different."
        return $null
    }

    return $orchestratorHost
}