function Read-WebConfigKey {
    param(
        [Parameter(Mandatory = $true)]
        [System.Xml.XmlDocument] $WebConfig,
        [Parameter(Mandatory = $true)]
        [string] $Key
    )

    $value = $WebConfig.SelectSingleNode("/configuration/appSettings/add[@key='$Key']").value

    if ([string]::IsNullOrWhiteSpace($value)) {
        return $value
    } else {
        $value.Trim()
    }
}

function Confirm-EncryptedSection {
    param(
        [Parameter(Mandatory = $true)]
        [System.Xml.XmlDocument] $WebConfig,
        [Parameter(Mandatory = $true)]
        [string] $Section,
        [Parameter(Mandatory = $false)]
        [System.Xml.XmlNamespaceManager]$NamespaceManager = $null
    )
    $encryptedSection = $WebConfig.SelectSingleNode("/configuration/$Section[@configProtectionProvider]", $NamespaceManager)
    return $encryptedSection -ne $null
}

function Write-ErrorMessage {
    param (
        [string] $Message
    )
    Write-Host "$message" -ForegroundColor Red
}

function Write-WarningMessage {
    param (
        [string] $Message
    )
    Write-Host "$message" -ForegroundColor Yellow
}

function Write-SuccessMessage {
    param (
        [string] $Message
    )
    Write-Host "$message" -ForegroundColor DarkGreen
}


function Write-ValidationResult {
    param(
        [Parameter(Mandatory = $true)]
        [ValidationResult] $ValidationResult,
        [Parameter(Mandatory = $false)]
        [string] $ValidationSubject,
        [Parameter(Mandatory = $false)]
        [bool] $OutputVerbose = $false
    )

    $message = "{0} validations: {1} succeeded, {2} failed and {3} warning(s)." `
        -f $ValidationSubject, $ValidationResult.Successes, $ValidationResult.Errors, $ValidationResult.Warnings
 
    if ($OutputVerbose) {
        Write-Verbose $message
    } else {
        Write-MessageWithValidation -Message $message -ValidationResult $ValidationResult
    }
}

function Write-MessageWithValidation {
    param(
        [Parameter(Mandatory = $true)]
        [string] $Message,
        [Parameter(Mandatory = $true)]
        [ValidationResult] $ValidationResult
    )
    
    if($ValidationResult.Errors -gt 0){
        Write-ErrorMessage $Message
    }
    elseif ($ValidationResult.Warnings -gt 0){
        Write-WarningMessage $Message
    }
    else{
        Write-SuccessMessage $Message
    }
}

function Invoke-SqlCommand {
    param(
        [Parameter(Mandatory = $true)]
        [string] $ConnectionString,
        [Parameter(Mandatory = $true)]
        [string] $Query)

    Use-Object($connection = new-object system.data.SqlClient.SQLConnection($ConnectionString)) {
        Use-Object($command = new-object System.Data.SqlClient.SqlCommand($Query,$connection)) {
            $connection.Open()

            Use-Object($adapter = New-Object System.Data.SqlClient.SqlDataAdapter $command) {
                $dataset = New-Object System.Data.DataSet
                $adapter.Fill($dataSet) | Out-Null
                return $dataSet.Tables
            }
        }
    }
}

function Use-Object
{
    [CmdletBinding()]
    param (
        [Parameter(Mandatory = $true)]
        [AllowEmptyString()]
        [AllowEmptyCollection()]
        [AllowNull()]
        [Object]
        $InputObject,
 
        [Parameter(Mandatory = $true)]
        [scriptblock]
        $ScriptBlock
    )
 
    try
    {
        . $ScriptBlock
    }
    finally
    {
        if ($null -ne $InputObject -and $InputObject -is [System.IDisposable])
        {
            $InputObject.Dispose()
        }
    }
}

