function Get-WebAppSite {
    param(
        [Parameter(Mandatory = $false)]
        [string] $SiteName,

        [Parameter(Mandatory = $false)]
        [string] $InstallationDirectory,

        [Parameter(Mandatory = $false)]
        [string] $WebApp
    )

    Write-Verbose "Checking $WebApp installation directory..."

    if ($InstallationDirectory -and $SiteName) {
        Write-WarningMessage "Site name and installation directory are both set. Using installation directory for $webapp installation discovery."
    }

    if ($InstallationDirectory) {
        if (Test-Path $InstallationDirectory) {
            Write-Verbose "Searching for $WebApp installation in '$InstallationDirectory' directory."
            $webAppSite = Get-Item "IIS:\Sites\*" -ErrorAction SilentlyContinue | Where-Object {$_.PhysicalPath -like $InstallationDirectory}
            if ($webAppSite -and $webAppSite.Count -eq 1) {
                Write-Verbose "$WebApp installation directory with site name '$($webAppSite.Name)' was found at '$($webAppSite.PhysicalPath)'."
                return $webAppSite
            }
            else {
                Write-Verbose "Could not find $WebApp installation with the specified installation directory '$InstallationDirectory'."
            }
        } else {
            Write-Verbose "$WebApp installation directory could not be found at '$InstallationDirectory'."
        }
    } elseif ($SiteName) {
        Write-Verbose "Discovering $WebApp installation directory..."

        $webAppSite = Get-Item "IIS:\Sites\$SiteName" -ErrorAction SilentlyContinue
        if ($webAppSite -and $webAppSite.count -eq 1) {
            Write-Verbose "$WebApp installation directory with site name '$SiteName' was found at '$($webAppSite.PhysicalPath)'."
            return $webAppSite
        } else {
            Write-Verbose "Could not find the $WebApp installation with the specified site name '$SiteName'."
        }
    } else {
        Write-Verbose "$WebApp site name or installation directory is required."
    }
}

function Get-WebAppVersion {
    param(
        [Parameter(Mandatory = $true)]
        [string] $WebAppBinPath,

        [Parameter(Mandatory = $true)]
        [string] $WebApp,

        [Parameter(Mandatory = $true)]
        [string] $WebAppDll
    )

    Write-Verbose "Discovering $WebApp version..."

    $webAppDllPath = Join-Path $WebAppBinPath $WebAppDll
    if (-not (Test-Path $webAppDllPath)) {
        return $null;
    }

    $webAppFileVersion = Get-ChildItem $webAppDllPath |
        Select-Object @{Name = "Version"; Expression = {$_.versionInfo}} |
        Select-Object -ExpandProperty Version |
        Select-Object -ExpandProperty FileVersion |
        Where-Object {-not [System.String]::IsNullOrWhiteSpace($_)}

    $majorMinorMatch = $webAppFileVersion | Select-String -Pattern "(\d+)\.(\d+)*"

    if($majorMinorMatch.Matches -and $majorMinorMatch.Matches[0].Success) {
        $groups = $majorMinorMatch.Matches[0].Groups

        $webAppVersion = [WebAppVersion]::new($groups[1].ToString(), $groups[2].ToString())

        Write-verbose "Current $WebApp version is $webAppVersion."
    } else {
        Write-WarningMessage "Could not find $WebApp version. Treat as the newest version."
        $webAppVersion = [WebAppVersion]::new(20,10)
    }
    
    return $webAppVersion
}

function Get-OrchestratorInstallation {
    param(
        [Parameter(Mandatory = $false)]
        [string] $SiteName,

        [Parameter(Mandatory = $false)]
        [string] $InstallationDirectory
    )

    $orchestratorSite = Get-WebAppSite `
        -SiteName $SiteName `
        -InstallationDirectory $InstallationDirectory `
        -WebApp "Orchestrator"
    
    if (-not $orchestratorSite) {
        return $null;
    }

    $binPath = Join-Path $orchestratorSite.PhysicalPath "bin"
    if (-not (Test-Path $binPath)) {
        $binPath = $orchestratorSite.PhysicalPath
    }

    $webAppDll = "UiPath.Orchestrator.Web.dll"
    $webAppDllPath = Join-Path $binPath $webAppDll
    if (-not (Test-Path $webAppDllPath)) {
        # Older versions of Orchestrator had different Web dll
        $webAppDll = "UiPath.Web.dll"
    }
    
    $orchestratorVersion = Get-WebAppVersion `
        -WebAppBinPath $binPath `
        -WebApp "Orchestrator" `
        -WebAppDll $webAppDll

    if(-not $orchestratorVersion){
        return $null;
    }

    return [WebAppInstallationInfo]::new($orchestratorSite, $orchestratorVersion, $binPath)
}

function Get-IdentitySite {
    param(
        [Parameter(Mandatory = $true)]
        [string] $OrchestratorSiteName
    )

    $identitySite = Get-Item "IIS:\Sites\$OrchestratorSiteName\Identity" -ErrorAction SilentlyContinue

    if ($identitySite -and $identitySite.Count -eq 1) {
        Write-Verbose "Identity site was found at 'IIS:\Sites\$OrchestratorSiteName\Identity'."
        return $identitySite
    }
    else {
        Write-Verbose "Identity site could not be found at 'IIS:\Sites\$OrchestratorSiteName\Identity'."
    }
}

function Test-OrchestratorLatestVersion {
    param(
        [Parameter(DontShow)]
        [WebAppInstallationInfo] $OrchestratorInstallation
    )

    if (-not $OrchestratorInstallation) {
        return $false
    }

    $latestVersion = [WebAppVersion]::new(20,10)

    return $latestVersion -eq $OrchestratorInstallation.Version 
}

function Get-WebAppConfig {
    param(
        [Parameter(Mandatory = $true)]
        [string] $WebAppInstallationDirectory,

        [Parameter(Mandatory = $true)]
        [string] $WebApp
    )
    
    if ($webApp -like "Orchestrator") {
        $appConfig = Get-WebConfig -InstallationDirectory $WebAppInstallationDirectory
    } elseif ($webApp -like "Identity") {
        $appConfig = Get-IdentityConfig -InstallationDirectory $WebAppInstallationDirectory
    }

    if (-not $appConfig.Config) {
        Write-ErrorMessage "Could not find $WebApp app config at '$($appConfig.Path)'"
    }

    return $appConfig.Config
}

function Get-WebConfig {
    param(
        [Parameter(Mandatory = $true)]
        [string] $InstallationDirectory
    )

    $webConfigPath = Join-Path $InstallationDirectory "Web.config"

    return Get-Config -ConfigPath $webConfigPath
}

function Get-IdentityConfig {
    param(
        [Parameter(Mandatory = $true)]
        [string] $InstallationDirectory
    )

    $identityConfigPath = Join-Path $InstallationDirectory "appsettings.Production.json"

    return Get-Config -ConfigPath $identityConfigPath
}

function Get-OrchestratorDllConfig {
    param(
        [Parameter(Mandatory = $true)]
        [string] $InstallationDirectory
    )

    $dllConfigPath = Join-Path $InstallationDirectory "UiPath.Orchestrator.dll.config"
    $appConfig = Get-Config -ConfigPath $dllConfigPath
    if (-not $appConfig.Config) {
        Write-ErrorMessage "Could not find UiPath.Orchestrator.dll.config at '$($dllConfigPath)'"
    }

    return $appConfig.Config
}

function Get-Config {
    param(
        [Parameter(Mandatory = $true)]
        [string] $ConfigPath 
    )

    if (Test-Path $ConfigPath) {
        $extension = (Get-Item $ConfigPath).Extension
        if ($extension -like ".json") {
            $config = Get-Content $ConfigPath | ConvertFrom-Json
        } elseif ($extension -like ".config") {
            $config = New-Object System.Xml.XmlDocument
            try {
                $config.Load($ConfigPath)
            } catch {
                $config = $null
            }
        }
    }

    return @{
        Config = $config;
        Path = $ConfigPath
    }
}