<#
    .SYNOPSIS
    Check the Orchestrator readiness pre and post 20.10 installation for several common errors.
    Update Orchestrator and Identity Server integration urls.
    Update UiPath Orchestrator SSL certificate or Identity token signing certificate.

    .DESCRIPTION
    Pre-Installation checks for UiPath Orchestrator platform:
        1. AspNetCoreModuleV2 installation
        2. Orchestrator CyberArk AIM Agent installation path is under Program Files(x86) directory if CyberArk is used in Orchestrator Web.config
        3. Orchestrator web.config section encryption
        4. Orchestrator configuration sections locked at server level
        5. Orchestrator host name, Identity(if installed) Config Orchestrator Url similarity with ssl certificate subject
        6. Orchestrator SqlServer Scaleout use
        7. Orchestrator external credential store plugins target framework compatibility with dotnet core runtime
        8. Orchestrator external Nlog plugins target framework to be compatibility with dotnet core runtime
        9. Orchestrator buckets root path should be added in the allowed list in the config file for any version starting 20.4. The connection string section needs to be available in config file
        10. The Identity server token signing certificate: exists, can be validated, has the appropriate key size, has a private key accessible to the app pool user

    Post-Installation checks for UiPath Orchestrator platform:
        1. Orchestrator host name, Identity Config Orchestrator Url similarity with ssl certificate subject
        2. Orchestrator buckets root path should be added in the allowed list in the config file. The connection string section needs to be available in config file
        3. Checks the Identity server token signing certificate: exists, can be validated, has the appropriate key size, has a private key accessible to the app pool user
        4. Inspect documentation page for post-installation checks for Orchestrator
    
    If any of the above checks is not valid, the script will output an error or an warning.

    UpdateUiPathUrl Command:
        1.Update UiPath.Orchestrator.dll.config OpenIdConnect and Identity integration related settings with the new Orchestrator Url and Identity Server Url.
        2.Update appsettings.Production.json(Identity Server config) with the new values for Orchestrator Url and Identity Server Url.
        3.Update the [Identity].[ClientRedirectUris] table with the new Orchestrator Url.
        4.SQLServer PowerShell module is required.
        5.Use this command with caution as it does not have roll-back mechanism if an error occurs during the invocation.
    
    UpdateUiPathCertificate Command:
        1.Update the Identity token signing certificate in appsettings.Production.json.
        2.Update the Orchestrator SSL certificate in IIS SSL bindings.
        3.Run Readiness command to validate the new certificates.

    LEGAL NOTICE: By installing and using this software, you (individual or legal entity) agree to the applicable LICENSE AGREEMENT available here: https://www.uipath.com/developers/all-editions/license-agreement. Please read it carefully. If you disagree with the license agreement, do not install or use the software and delete it from your computer.

    .PARAMETER Help
    Option to display the help for a specific invocation command or option.

    .PARAMETER Parameter
    The parameter name to display help.

    .PARAMETER Readiness
    Option to invoke readiness validation pre and post installation of Orchestrator 20.10. This option will discover the current Orchestrator installation and it will call pre or post installation checks accordingly.
    Compatible with -Help option.

    .PARAMETER SiteName
    Input parameter for for Readiness, UpdateUiPathUrl and UpdateUiPathCertificate commands.
    The name of the Orchestrator site that can be found in IIS on the target machine. Usually this can be found as 'UiPath Orchestrator'.

    .PARAMETER InstallationDirectory
    Path to the Orchestrator installation. Usually Orchestrator installation can be found at 'C:\Program Files (x86)\UiPath\Orchestrator'.

    .PARAMETER UpdateUiPathUrl
    Option to update Orchestrator and Identity Server integration urls. This command will update UiPath.Orchestrator.dll.config and appsettings.Production.json configuration file for Orchestrator and Identity Server respectively.
    Invoking this command will also update the [Identity].[ClientRedirectUris] table with the new Orchestrator Url.
    SQLServer Powershell module is required.
    Use this command with caution as it does not have roll-back mechanism if an error occurs during the invocation.

    .PARAMETER OrchestratorUrl
    Input parameter for UpdateUiPathUrl command.
    The new Orchestrator url used by Identity Server in authentication flows.

    .PARAMETER IdentityServerUrl
    Input parameter for UpdateUiPathUrl command. If it is not set the default value is '{OrchestratorUrl}/Identity'.
    The new Identity Server url used by Orchestrator in authentication flows.

    .PARAMETER SqlConnectionString
    Input parameter for UpdateUiPathUrl command. If it is not set the value is read from appsettings.Production.json(Identity Server config).
    The connection string to the UiPath database.

    .PARAMETER UpdateUiPathCertificate
    Option to update the Orchestrator SSL certificate and Identity Server token signing certificate.
    It is recommended to run Readiness command to validate the new certificates.

    .PARAMETER NewTokenSigningThumbprint
    Input parameter for UpdateUiPathCertificate. The computer certificate store thumbprint of the new token signing certificate.
    If it is set it will update appsettings.Production.json(Identity Server config) with the new certificate thumbprint.

    .PARAMETER NewSSLThumbprint
    Input parameter for UpdateUiPathCertificate. The computer certificate store thumbprint of the new Orchestrator SSL certificate
    If it is set it will update the Orchestrator IIS ssl bindings with the new certificate thumbprint.

    .PARAMETER $KeepOldCertificate
    Flag indicating if current certificate should continue to be used as a validation key for tokens that were signed with it.
#>

#Requires -RunAsAdministrator
#Requires -PSEdition Desktop

Using module .\Modules\Models.psm1

[CmdletBinding(DefaultParameterSetName = "Help")]
param(
    [Parameter(Mandatory = $false, ParameterSetName = "Help")]
    [switch]$Help,

    [Parameter(Mandatory = $false, ParameterSetName = "Help")]
    [string]$Parameter,

    [Parameter(Mandatory = $false, ParameterSetName = "Help")]
    [Parameter(Mandatory = $false, ParameterSetName = "ReadinessSiteName")]
    [Parameter(Mandatory = $false, ParameterSetName = "ReadinessInstallationDirectory")]
    [switch]$Readiness,

    [Parameter(Mandatory = $true, ParameterSetName = "ReadinessSiteName")]
    [Parameter(Mandatory = $true, ParameterSetName = "UpdateUiPathUrl")]
    [Parameter(Mandatory = $true, ParameterSetName = "UpdateUiPathCertificate")]
    [string] $SiteName = "UiPath Orchestrator",

    [Parameter(Mandatory = $true, ParameterSetName = "ReadinessInstallationDirectory")]
    [string]$InstallationDirectory,

    [Parameter(Mandatory = $false, ParameterSetName = "Help")]
    [Parameter(Mandatory = $true, ParameterSetName = "UpdateUiPathUrl")]
    [switch]$UpdateUiPathUrl,

    [Parameter(Mandatory = $true, ParameterSetName = "UpdateUiPathUrl")]
    [string] $OrchestratorUrl,

    [Parameter(Mandatory = $false, ParameterSetName = "UpdateUiPathUrl")]
    [string] $IdentityServerUrl,

    [Parameter(Mandatory = $false, ParameterSetName = "UpdateUiPathUrl")]
    [string] $SqlConnectionString,

    [Parameter(Mandatory = $false, ParameterSetName = "Help")]
    [Parameter(Mandatory = $true, ParameterSetName = "UpdateUiPathCertificate")]
    [switch]$UpdateUiPathCertificate,

    [Parameter(Mandatory = $false, ParameterSetName = "UpdateUiPathCertificate")]
    [string] $NewTokenSigningThumbprint,

    [Parameter(Mandatory = $false, ParameterSetName = "UpdateUiPathCertificate")]
    [string] $NewSSLThumbprint,

    [Parameter(Mandatory = $false, ParameterSetName = "UpdateUiPathCertificate")]
    [bool] $KeepOldCertificate = $true
)

Import-Module "$PSScriptRoot\Modules\Utils.psm1" -Force -Verbose:$false
Import-Module "$PSScriptRoot\Modules\WebApplication" -Force  -Verbose:$false
Import-Module "$PSScriptRoot\Modules\Orchestrator.Checks.psm1" -Force -Verbose:$false
Import-Module "$PSScriptRoot\Modules\Identity.Checks.psm1" -Force -Verbose:$false
Import-Module "$PSScriptRoot\Modules\EnvironmentOperations.psm1" -Force -Verbose:$false
Import-Module WebAdministration -Verbose:$false

function Main {
    if ($Help -and $Readiness) {
        Get-Help Test-PlatformReadiness
        return
    } elseif ($Help -and $UpdateUiPathUrl) {
        Get-Help Update-UiPathUrl
        return
    } elseif ($Help -and $UpdateUiPathCertificate) {
        Get-Help Update-UiPathCertificate
        return
    } elseif ($Help -or ($script:PSBoundParameters.ContainsKey("Parameter"))) {
        Get-Help -Name $myInvocation.ScriptName -Parameter $script:PSBoundParameters["Parameter"]
    } elseif ($Help -or ($script:PSBoundParameters.count -eq 0)) {
        Get-Help -Name $myInvocation.ScriptName
        return
    }

    if ($Readiness) {
        if ($script:PSBoundParameters.ContainsKey('InstallationDirectory')) {
            Test-PlatformReadiness -InstallationDirectory $InstallationDirectory
        } else {
            Test-PlatformReadiness -SiteName $SiteName
        }
    } elseif ($UpdateUiPathUrl) {
        Update-UiPathUrl `
            -SiteName $SiteName `
            -OrchestratorUrl $OrchestratorUrl `
            -IdentityServerUrl $IdentityServerUrl `
            -SqlConnectionString $SqlConnectionString
    } elseif ($UpdateUiPathCertificate) {
        Update-UiPathCertificate `
            -SiteName $SiteName `
            -NewTokenSigningThumbprint $NewTokenSigningThumbprint `
            -NewSSLThumbprint $NewSSLThumbprint `
            -KeepOldCertificate $KeepOldCertificate
    }
}

<#
    .SYNOPSIS
        Check the Orchestrator readiness pre and post 20.10 installation for several common errors.
    .DESCRIPTION
        Pre-Installation checks for UiPath Orchestrator platform:
            1. AspNetCoreModuleV2 installation
            2. Orchestrator CyberArk AIM Agent installation path is under Program Files(x86) directory if CyberArk is used in Orchestrator Web.config
            3. Orchestrator web.config section encryption
            4. Orchestrator configuration sections locked at server level
            5. Orchestrator host name, Identity(if installed) Config Orchestrator Url similarity with ssl certificate subject
            6. Orchestrator SqlServer Scaleout use
            7. Orchestrator external credential store plugins target framework compatibility with dotnet core runtime
            8. Orchestrator external Nlog plugins target framework to be compatibility with dotnet core runtime
        9. Orchestrator buckets root path should be added in the allowed list in the config file for any version starting 20.4. The connection string section needs to be available in config file
            10. The Identity server token signing certificate: exists, can be validated, has the appropriate key size, has a private key accessible to the app pool user

        Post-Installation checks for UiPath Orchestrator platform:
            1. Orchestrator host name, Identity Config Orchestrator Url similarity with ssl certificate subject
        2. Orchestrator buckets root path should be added in the allowed list in the config file. The connection string section needs to be available in config file
            3. Checks the Identity server token signing certificate: exists, can be validated, has the appropriate key size, has a private key accessible to the app pool user
            4. Inspect documentation page for post-installation checks for Orchestrator
        
        If any of the above are false, the script will output an error or an warning

    .PARAMETER SiteName
    The name of the Orchestrator site that can be found in IIS on the target machine. Usually this can be found as 'UiPath Orchestrator'.

    .PARAMETER InstallationDirectory
    Path to the Orchestrator installation. Usually Orchestrator installation can be found at 'C:\Program Files (x86)\UiPath\Orchestrator'.
#>

function Test-PlatformReadiness {
    [CmdletBinding(DefaultParameterSetName='Site Name')]
    param(
        [Parameter(Mandatory = $true, ParameterSetName = "Site Name")]
        [string]$SiteName,

        [Parameter(Mandatory = $true, ParameterSetName = "Installation Directory")]
        [string]$InstallationDirectory
    )

    $orchestratorInstallation = Get-OrchestratorInstallation -SiteName $SiteName -InstallationDirectory $InstallationDirectory

    if (-not (Test-OrchestratorLatestVersion -OrchestratorInstallation $orchestratorInstallation)) {
        Test-PlatformPreInstallation -OrchestratorInstallation $orchestratorInstallation
    } else {
        Test-PlatformPostInstallation -OrchestratorInstallation $orchestratorInstallation
    }
}

function Test-PlatformPreInstallation() {
    param(
        [Parameter(Mandatory = $false)]
        [string]$SiteName,

        [Parameter(Mandatory = $false)]
        [string]$InstallationDirectory,

        [Parameter(DontShow)]
        [WebAppInstallationInfo] $OrchestratorInstallation
    )

    Write-Host "Validating 20.10 pre-installation requirements..."

    $validationResult = [ValidationResult]::new()

    # Check environment requirements. Orchestrator does not have to be installed.
    $validationResult += Test-WebCoreHostModuleInstallation

    # Check Orchestrator installation
    if (-not $PSBoundParameters.ContainsKey('OrchestratorInstallation')) {
        $OrchestratorInstallation = Get-OrchestratorInstallation -SiteName $SiteName -InstallationDirectory $InstallationDirectory
    }

    if (-not $OrchestratorInstallation) {
        Write-ErrorMessage "Orchestrator is not installed. Can not perform the rest of platform pre-installation checks."
        $validationResult.Errors++
        Write-ValidationResult $validationResult -ValidationSubject "Platform readiness"
        return
    } elseif (Test-OrchestratorLatestVersion -OrchestratorInstallation $OrchestratorInstallation) {
        Write-WarningMessage "Latest Orchestrator version is already installed. Can not perform the rest of platform pre-installation checks."
        $validationResult.Warning++
        Write-ValidationResult $validationResult -ValidationSubject "Platform readiness"
        return
    }

    # Orchestrator is installed.
    $webConfig = Get-WebAppConfig `
        -WebAppInstallationDirectory $OrchestratorInstallation.Site.PhysicalPath `
        -WebApp "Orchestrator"

    if (-not $webConfig) {
        # Orchestrator web.config does not exist.
        Write-ErrorMessage "Could not complete the rest of platform pre-installation checks."
        $validationResult.Errors++
        Write-ValidationResult $validationResult -ValidationSubject "Platform readiness"
        return
    }

    # Check Orchestrator requirements. Orchestrator has to be installed.
    $validationResult += Test-OrchestratorCyberArkAIMInstallation -WebConfig $webConfig

    $validationResult += Test-OrchestratorWebConfigEncryption -WebConfig $webConfig

    $validationResult += Test-OrchestratorWebConfigLockedSections

    $identitySite = Get-IdentitySite -OrchestratorSiteName $OrchestratorInstallation.Site.Name
    if ($identitySite) {
        $identityConfig = Get-WebAppConfig `
        -WebAppInstallationDirectory $IdentitySite.PhysicalPath `
        -WebApp "Identity"
    }

    $validationResult += Test-OrchestratorCertificateSubject -IdentityConfig $identityConfig -OrchestratorSiteName $OrchestratorInstallation.Site.Name

    $validationResult += Test-OrchestratorSqlServerScaleoutUse -WebConfig $webConfig

    $validationResult += Test-OrchestratorCredentialStoresPlugins -WebConfig $webConfig -OrchestratorDirectory $OrchestratorInstallation.Site.PhysicalPath

    $validationResult += Test-OrchestratorNLogPlugins -WebConfig $webConfig -OrchestratorBinPath $OrchestratorInstallation.BinPath
    
    $bucketsVersion = [WebAppVersion]::new(20, 4)
    if ($OrchestratorInstallation.Version -ge $bucketsVersion) {
        $validationResult += Test-OrchestratorFileSystemBuckets -AppConfig $webConfig
    }

    if (-not $identitySite) {
        Write-ErrorMessage "Identity is not installed. Can not complete the Identity pre-installation checks."
        $validationResult.Errors++
        Write-ValidationResult $validationResult -ValidationSubject "Platform readiness"
        return
    }

    if (-not $identityConfig) {
        Write-ErrorMessage "Can not complete the Identity pre-installation checks."
        $validationResult.Errors++
        Write-ValidationResult $validationResult -ValidationSubject "Platform readiness"
        return
    }

    # Check Identity requirements. Identity has to be installed.
    $validationResult += Test-PlatformCertificates `
        -IdentitySite $IdentitySite `
        -IdentityConfig $identityConfig `
        -OrchestratorSiteName $OrchestratorInstallation.Site.Name

    Write-Host "All 20.10 pre-installation checks are done."

    Write-ValidationResult $validationResult -ValidationSubject "Platform readiness"
}

function Test-PlatformPostInstallation {
    param(
        [Parameter(Mandatory = $false)]
        [string]$SiteName,

        [Parameter(Mandatory = $false)]
        [string]$InstallationDirectory,
        
        [Parameter(DontShow)]
        [WebAppInstallationInfo] $OrchestratorInstallation
    )

    Write-Host "Validating 20.10 post-installation requirements..."

    $validationResult = [ValidationResult]::new()

    # Check Orchestrator installation
    if (-not $PSBoundParameters.ContainsKey('OrchestratorInstallation')) {
        $OrchestratorInstallation = Get-OrchestratorInstallation -SiteName $SiteName -InstallationDirectory $InstallationDirectory
    }

    if (-not $OrchestratorInstallation) {
        Write-ErrorMessage "Orchestrator is not installed. Can not perform the rest of platform post-installation checks."
        $validationResult.Errors++
        Write-ValidationResult $validationResult -ValidationSubject "Platform readiness"
        return
    } elseif (-not (Test-OrchestratorLatestVersion -OrchestratorInstallation $OrchestratorInstallation)) {
        Write-WarningMessage "Latest Orchestrator version, 20.10, is not installed. Can not perform the rest of platform pre-installation checks."
        $validationResult.Warning++
        Write-ValidationResult $validationResult -ValidationSubject "Platform readiness"
        return
    }
    # Orchestrator is installed.
    $appConfig = Get-OrchestratorDllConfig -InstallationDirectory $OrchestratorInstallation.Site.PhysicalPath
    if (-not $appConfig) {
        # Orchestrator app.config does not exist.
        Write-ErrorMessage "UiPath.Orchestrator.dll.config was not found. Can not complete the rest of the post-installation checks."
        $validationResult.Errors++
        Write-ValidationResult $validationResult -ValidationSubject "Platform readiness"
        return
    }

    $validationResult += Test-OrchestratorFileSystemBuckets -AppConfig $appConfig

    $identitySite = Get-IdentitySite -OrchestratorSiteName $OrchestratorInstallation.Site.Name

    if (-not $identitySite) {
        Write-ErrorMessage "Identity is not installed. Can not complete the Identity post-installation checks."
        $validationResult.Errors++
        Write-ValidationResult $validationResult -ValidationSubject "Platform readiness"
        return
    }


    $identityConfig = Get-WebAppConfig `
        -WebAppInstallationDirectory $IdentitySite.PhysicalPath `
        -WebApp "Identity"

    if (-not $identityConfig) {
        Write-ErrorMessage "Can not complete the Identity post-installation checks."
        $validationResult.Errors++
        Write-ValidationResult $validationResult -ValidationSubject "Platform readiness"
        return
    }

    $validationResult += Test-OrchestratorCertificateSubject `
        -IdentityConfig $identityConfig `
        -OrchestratorSiteName $OrchestratorInstallation.Site.Name

    # Check Identity requirements. Identity has to be installed.
    $validationResult += Test-PlatformCertificates `
        -IdentitySite $IdentitySite `
        -IdentityConfig $identityConfig `
        -OrchestratorSiteName $OrchestratorInstallation.Site.Name

    Write-Host "All 20.10 post-installation checks are done. Inspect post-installation requirements at https://docs.uipath.com/installation-and-upgrade/v2020.10/docs/post-installation-considerations."
    Write-ValidationResult $validationResult -ValidationSubject "Platform readiness"
}

Main